package Schedule;

require 5.001;

my $start_time = time;

use strict;

use CWorker;
use MicroTime;
use vars qw( $profile );

$profile = 0;


sub schedule {
    my( $objs, $time, @args ) = @_;

    my $obj;
    my $minute = $time;
    my $old_niceness = 0;
    my $nextevent = time+60;
    my $currenttime = time();
    foreach $obj (sort { $a->niceness() <=> $b->niceness() } @$objs) {
	Platform::poll();
	my $niceness = $obj->niceness();
	if( $niceness != $old_niceness ) {
	    CWorker::do_them();
	    $old_niceness = $niceness;
	}
        my $time;
	if( $profile ) {
	    $time = MicroTime::gettimeofday();
	}
	$obj->sched_run( $minute, $currenttime, @args );
	if( $profile ) {
	    $time = MicroTime::difftime( $time );
	    if( $time > 0.5 ) {
	        BigSister::common::log( "info", "profile: Scheduler spent ".(int($time*100)/100)."s in $obj\n" );
	    }
	}
	if( $obj->{"sched_nextdue"} 
	    && ($obj->{"sched_nextdue"} < $nextevent) ) {
	    $nextevent = $obj->{"sched_nextdue"};
	}
	CWorker::one_pass();
    }
    CWorker::one_pass();
    return $nextevent;
}


sub new
{
    my ($this, $frequency, $procedure) = @_;
    my  $class = ref($this) || $this;
    my  $self = {};

    bless($self, $class);
    $self->sched_init( $frequency, $procedure );

    return($self);
}


sub sched_init {
    my ($self, $frequency, $procedure) = @_;
    
    ($frequency = 5) unless( $frequency );
    $self->{"sched_frequency"} = $frequency;
    ($self->{"sched_procedure"} = $procedure) if( $procedure );
    $self->{"sched_nextdue"} = 0;
    $self->{"sched_currenttime"} = -1;
}


sub sched_run {
    my( $self, $minute, $currenttime, @args ) = @_;

    $self->{"sched_minute"} = $minute;
    unless( $self->{"sched_frequency"} ) {
	$self->sched_init();
    }
    return if( $self->{"sched_frequency"} eq "none" );
    return if( $minute && ($minute < $self->{"sched_nextdue"})
        && ! ($minute == $self->{"sched_currenttime"}) );
    $self->{"sched_currenttime"} = $minute;
    if( $self->{"sched_nextdue"} <= $minute ) {
        $self->{"sched_nextdue"} += $self->{"sched_frequency"}*60;
    }
    if( ($self->{"sched_nextdue"} <= $minute) || ! $minute ) {
        # skip runs until obj time > minute
        $self->{"sched_nextdue"} =
	    (int( (($minute ? $minute : $currenttime) - $start_time ) / 60 / $self->{"sched_frequency"} ) + 1) * 60 * $self->{"sched_frequency"} + $start_time; 
    }
    return( $self->run( @args ) );
}


sub run {
    my( $self, @args ) = @_;

    if( $self->{"sched_procedure"} ) {
	$self->{"sched_procedure"}->( @args );
    }
}



sub sched_frequency {
    my( $self, $frequency ) = @_;

    ($frequency = 5) unless( $frequency );
    $self->{"sched_frequency"} = $frequency;
}



sub sched_procedure {
    my( $self, $procedure ) = @_;

    $self->{"sched_procedure"} = $procedure;
}


sub sched_gettime {
    my( $self ) = @_;
    
    (defined $self->{"sched_minute"}) || return(-1);
    return( $self->{"sched_minute"} );
}


sub niceness {
    return 0;
}

	     
1;
