package Requester::expedap;

use strict;
use Requester::SlowRequester;

our(@ISA);
@ISA = qw(Requester::SlowRequester);

my $expedap = new Requester::expedap();

sub new
{
   my( $this, @args ) = @_;
   my $class = ref($this) || $this;
   my $self = {};
   bless( $self, $class );
   $self->register("expedap");	 
   return( $self );
}

sub startup {
  my ($self, @args) = @_;
	my $sysPath = $ENV{PATH};
	Platform::mergepath( $BigSister::common::fs{"adm"}."/expedap", 0);
	print "search path is: $ENV{PATH}\n" if ($BigSister::common::dl > 4);
	my $fh = $self -> SUPER::startup(@args);
	$ENV{PATH} = $sysPath;
	print "search path is reset to: $ENV{PATH}\n" if ($BigSister::common::dl > 4);
	return $fh;
}

sub getinfo {
  my( $self, $vars, $indexes ) = @_;

  # we cannot walk the expedap domain
  return undef unless( $indexes );
	return undef unless ( $vars );
	
	my $cmd = $vars -> [0];
	$cmd =~ s/\..*$//;
	
	# calculate arguments
	my $args = $self -> {"env"} -> {"args"};
	
	# if argument_list is set, then positional parameters are used, else all
	# parameters in $args are passed als named parameters
	my $argument_list = $self -> {"env"} -> expand_values( "argument_list" );
	print "argument_list: @$argument_list\n" if ($BigSister::common::dl > 4);
	if ( scalar @$argument_list > 0) {
	  #positional arguments
		foreach my $key ( @$argument_list ) {
			 $cmd .= " ".$args -> { $key } if ( exists $args -> { $key } );
		}
	} else {
		foreach my $key ( keys %$args ) {
			$cmd .= " $key=".$args -> { $key };
		}
	}
	
	$self -> command( $cmd );
	print "Expedap command is: $cmd\n" if ($BigSister::common::dl > 4);
	
	# the slowrequester object
	my $srq = $self -> SUPER::getinfo($vars, $indexes);
	
	# pass the baseId to the slowrequester
	$srq->{"Expedap::baseId"} = $self->{"env"}->expand_value( "id" );
	print "getinfo baseId: ".$srq->{"Expedap::baseId"}."\n" if ($BigSister::common::dl > 4);
	
	return $srq;
}


sub do {
	my ($self, $srq) = @_;
	return unless (my $line = $srq -> readln());
	
	
	while ( $line ) {
		chomp($line);

		if ( $line =~ /END/ ) {
			$srq -> status(3, "finished OK");
			return;
		}
		 
		if ( $line =~ /(.*)\.(\d+)\s*=\s*(.*)/ ) {
			print "test result: $line\n" if ($BigSister::common::dl > 4);
			
			my $baseId = $srq->{"Expedap::baseId"};
			print "do baseId: $baseId\n" if ($BigSister::common::dl > 6);
			my $id = $baseId + $2;
			
			my $info = $srq -> {"Test.info"};
			$info -> { $id } -> { $1 } = $3;
			print "info -> { $id } -> { $1 } = $3\n" if ($BigSister::common::dl > 6);
		}

		$line = $srq -> readln();
	}	
}

1;
